from .doc import doc
from .fig import adaptive_figsize, norm_arr, \
    set_ticks, set_spines, color_mapper_cat, \
    color_mapper_val, mask_triu, get_cmap_colors, \
    create_cmap, get_render_size
from .geo import rotate_points, normalize
from .params import set_default
